
/*  Name:          MAIN32v641   ****Copy before editing****
    By:            Bob Found, Indian Harbour, NS Canada
	  Start Date:    12Jun2018
    Last Rev:      18Apr2019
	  MCU:           Arduino UNO

    V6.41: added print statements to each function to display on serial monitor.  Removed HappyFace.
    Description: Main program for Christmas Display Star, 7 foot diameter. Programming started on small desktop model,
    full size programming began at version 2.4.  Full size colors are more vivid than those on model.

	  H/W:  7-MCP23017 port expanders, custom PCBs, 7-segment model of 16 RGB LEDs each  Model uses ULN drivers (sink)
          but full size project uses UDN2981 drivers (source).   Also had to use I2C extenders for full size.

    Colors: red, green, blue, magenta, cyan, yellow, white
    Segments: 5 points with one center pentagon that consists of TWO segments, for total 7 segments
    Each segment has custom I/O port expander board and I2C line extenders.  Pentagon uses one extender for both segments
*/
//*********************Libraries and Global Variables*****************************
#include <Centipede.h>//Centipede has best variables for writing to port expanders
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <elapsedMillis.h>
#include <LiquidCrystal_PCF8574.h>
LiquidCrystal_PCF8574 lcd(0x27);
Centipede CS;
// Note:  Colors don't come out correctly on model, but work okay on full size
const byte pins[3] = {9, 10, 11};
const byte Black = 0x00;
const byte Red = 0x02; //Red bus, pin 9 on UNO  eg use PORTB=Red 0000 0010
const byte Green = 0x04; //Green bus, pin 10 on UNO  0000 0100
const byte Yellow = 0x06; //0000 0110
const byte Blue = 0x08; //Blue bus, Pin 11 on UNO  0000 1000
const byte Magenta = 0x0A;//0000 1010
const byte Cyan = 0x0C;//0000 1100
const byte White = 0x0E;//0000 1110
const byte rgb[7] = {0x02, 0x04, 0x08, 0x0A, 0x06, 0x0C, 0x0E}; // LED colors 02=red, 04=green, 08=blue
const int RedPin = 9;// These 3 pins are for analog PWM (same as DIO pins, but change function)
const int GreenPin = 10;
const int BluePin = 11;
const int off = 0x0000;
const int allon = 0xFFFF;
byte segment;
int breathe_delay = 5;   // delay between loops
unsigned long breathe_time = millis();
int count;
byte rep;
static float in = 4.712;//used for fading/breath
float out;
int del;
//***Order of passing arguments is alphabetical: del, color, segment, rep, runtime***
//*****************************Declare all arrays here****************************
int InsideArray[5]; //int relates to values stored in array
int OutsideArray[7];
int SolidArray[7]; //Each segment all lights on
int BlackHoleArray[3];
int SideArray[5];
int ForSeqArray[15];
int RevSeqArray[15];
int PointsPairsArray[7];
int SweepArray20[7];
int SweepArray2122[5];
int SweepArray2324[5];

const int SweepArray[15] = {0x03C0, 0x0070, 0x0420, 0x3000, 0x0088, 0x0810,
                            0x4800, 0x0104, 0x1008, 0x0400, 0x0202, 0x2004, 0x0001, 0x4002, 0x0001
                           };//this array is for pentagon only0xC800 lights very center light, 0x4800 does not
byte SineArray[32] = {  0,   1,   2,   3,   4,   5,   7,   9, 12,  15,  18,  22,  27,  32,  38,  44,
                        51,  58,  67,  76,  86,  96, 108, 120, 134, 148, 163, 180, 197, 216, 235, 255
                     };//Keep for now...may need later

const byte PortArray[15] = {5, 6, 5, 6, 6, 5, 6, 6, 5, 6, 6, 5, 6, 5, 5};//pentagon segments, 5=outer, 6=inner
const byte StarSwapColorArray[7]={Red,Blue,Red, Red, Blue,Blue, Blue};
const byte ColorArray1[7] = {Red, Green, Blue, Magenta, Cyan, Yellow, White};
const byte ColorArray2[7] = {White, Red, Green, Blue, Magenta, Cyan, Yellow};
const byte ColorArray3[7] = {Yellow, White, Red, Green, Blue, Magenta, Cyan};
const byte ColorArray4[7] = {Cyan, Yellow, White, Red, Green, Blue, Magenta};
const byte ColorArray5[7] = {Magenta, Cyan, Yellow, White, Red, Green, Blue};
const byte ColorArray6[7] = {Blue, Magenta, Cyan, Yellow, White, Red, Green};
const byte ColorArray7[7] = {Green, Blue, Magenta, Cyan, Yellow, White, Red};

const int DelayArray[10] = {0x100, 0x90, 0x80, 0x70, 0x60, 0x50, 0x40, 0x30, 0x20, 0x10};//decreasing amounts of delay=speed up
const int DelayArrayFast[10] = {0x50, 0x45, 0x40, 0x35, 0x30, 0x25, 0x20, 0x15, 0x10, 0x05};
const byte CounterRotArray[28] = {5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6,//pentagon at 5 and 6 rotate opposite ways
                                  5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6
                                 };
const int TipsArray[10] = {5, 21, 37, 53, 69, 80, 83, 86, 89, 92};//tips of segments, points and pentagon
/*void SetArray() {//this is array created for Swarm() Uses shift bit left
  for (i = 0; i < 15; i++) //create the array
    byte z = 1 << (i); //shift bit left
  ForSeqArray[i] = z;*/

//**************************  Setup  ****************************************
void setup() {
  Wire.begin(); // start I2C
  Wire.beginTransmission(0x27);
  Serial.begin(9600);
  CS.initialize();//Centipede uses single digit for addressing, eg 0 instead of 20, 4 instead of 24
  for (byte segment = 0; segment < 8; segment++)   //Init all 7 I2C devices
  {
    CS.portMode(segment, 0x00);
  }
  lcd.begin(20, 4);
  //lcd.display();
  lcd.setCursor (0, 0);
  lcd.print(F(" Found Engineering"));//F invokes function to keep text in program memory
  // initialize the DIO pins as outputs
  for (byte i = 0; i < 3; i++)
  { pinMode(pins[i], OUTPUT);
  }

  //*******************Place all arrays here**********************************
  BlackHoleArray[0] = 0x7FFF;  // all outside pentagon
  BlackHoleArray[1] = 0x03FF;  // middle pentagon
  BlackHoleArray[2] = 0x7C00; // center pentagon

  InsideArray[0] = 0x0070;  // top of pentagon, 3 lights    * * *
  InsideArray[1] = 0x0088;  // next two down.              *     *
  InsideArray[2] = 0x0104;  //                            *       *
  InsideArray[3] = 0x0202;  //                              *   *
  InsideArray[4] = 0x0001;   //                               *
  InsideArray[5] = 0x3000;  //Center of pentagon, 5 lights only top two   * *
  InsideArray[6] = 0x4800;  //second two, left right                     *   *
  InsideArray[7] = 0x0400;  //bottom point                                 *

  OutsideArray[0] = 0x03C0;  // top of pentagon, 4 lights         * * * *
  OutsideArray[1] = 0x0420;  // next two down.                   *       *
  OutsideArray[2] = 0x0810;  //                                 *         *
  OutsideArray[3] = 0x1008;//                                  *           *
  OutsideArray[4] = 0x2004;//                                    *       *
  OutsideArray[5] = 0x4002;//                                      *   *
  OutsideArray[6] = 0x0001; // bottom point of pentagon              *//

  SideArray[0] = 0x03C0;  // top of pentagon, 4 lights
  SideArray[1] = 0x0038;  // top left  see previous
  SideArray[2] = 0x1E00; // top right see previous
  SideArray[3] = 0x000F; // bottom left  etc
  SideArray[4] = 0x7001; // bottom right etc

  PointsPairsArray[0] = 0x0020; //peak of point  *    Can be used for all points, no matter what address
  PointsPairsArray[1] = 0x0050; //next two      * *
  PointsPairsArray[2] = 0x0088; //             *   *
  PointsPairsArray[3] = 0x0104; //            *     *
  PointsPairsArray[4] = 0x0202; //           *       *
  PointsPairsArray[5] = 0x0401; //          *         *
  PointsPairsArray[6] = 0xF800; //         * * * * * * *

  ForSeqArray[0] = 0x0001;  //can also create this array using shift bit function (void SetArray)
  ForSeqArray[1] = 0x0002;
  ForSeqArray[2] = 0x0004;
  ForSeqArray[3] = 0x0008;
  ForSeqArray[4] = 0x0010;
  ForSeqArray[5] = 0x0020;
  ForSeqArray[6] = 0x0040;
  ForSeqArray[7] = 0x0080;
  ForSeqArray[8] = 0x0100;
  ForSeqArray[9] = 0x0200;
  ForSeqArray[10] = 0x0400;
  ForSeqArray[11] = 0x0800;
  ForSeqArray[12] = 0x1000;
  ForSeqArray[13] = 0x2000;
  ForSeqArray[14] = 0x4000;

  RevSeqArray[0] = 0x4000;//Should be able to use ForSeqArray, and call elements in reverse, eliminating this array.
  RevSeqArray[1] = 0x2000;
  RevSeqArray[2] = 0x1000;
  RevSeqArray[3] = 0x0800;
  RevSeqArray[4] = 0x0400;
  RevSeqArray[5] = 0x0200;
  RevSeqArray[6] = 0x0100;
  RevSeqArray[7] = 0x0080;
  RevSeqArray[8] = 0x0040;
  RevSeqArray[9] = 0x0020;
  RevSeqArray[10] = 0x0010;
  RevSeqArray[11] = 0x0008;
  RevSeqArray[12] = 0x0004;
  RevSeqArray[13] = 0x0002;
  RevSeqArray[14] = 0x0001;

  SweepArray20[0] = 0x0020; //peak of point  *     Address 20
  SweepArray20[1] = 0x0050; //next two      * *
  SweepArray20[2] = 0x0088; //             *   *
  SweepArray20[3] = 0x0104; //            *     *
  SweepArray20[4] = 0x0202; //           *       *
  SweepArray20[5] = 0x0401; //          *         *
  SweepArray20[6] = 0xF800; //         * * * * * * *

  SweepArray2122[0] = 0x0FE0;//Addr 21 AND 22 top side
  SweepArray2122[1] = 0x1020;// apex and next down
  SweepArray2122[2] = 0x2020;
  SweepArray2122[3] = 0x4020;
  SweepArray2122[4] = 0x803F;// bottom side

  SweepArray2324[0] = 0x803F;//Addr 23 and 24 right or left side  (Reverse of 2122)
  SweepArray2324[1] = 0x4020;
  SweepArray2324[2] = 0x2020;
  SweepArray2324[3] = 0x1020;
  SweepArray2324[4] = 0x0FE0;// bottom side
}
 unsigned long previousMillis=0;
//****************************************************************************
//***********************  MAIN PROGRAM  *************************************
//****************************************************************************
void loop()
{  unsigned long currentMillis = millis(); //uncomment if running Timer.
  
    DispVolts();
  //******************STAR FUNCTIONS START HERE******************************
  AllEdges1();//formally PentagonEdges, now all segment edges. 4 light combos flicker in each segment, change color KEEP
  AllOnSeq(Magenta);//  all segments on simultaneously in CW in color, slowly blink out KEEP
  AllLightSequence(15);//every light momentarily on in sequence 0-111, in all colors in sequence RGB...Pass delay value
  BlackHole(Blue, 4);//all segments from apex of points to center pentagon ok red for points, then yellow and white for pentagon pass reps
  BlinkSpinAccel();//blinks each segment in each rgb color. Uses faster delay array.  Use once.
  //Breathe2(5, RedPin,3); //Keep separate.  Combined both in BreathePentagon()???  Don't use together
  //Breathe2(6, BluePin,3);//(segment, colpin, rep)
  FadeAll(Red,4);//all 110 lights fading at same time. Fades passed color with color in FadeAll()Rep=4 internal  change color
  FadeAllPoints(Blue, 1); //turn on each individual point, fade with different colors.  
  FadePentagons(Blue, 3);//fades both pentagons.  Starts bright, fadesout.  Fade not obvious if colors mixed.  KEEP
  FillIn(50, Magenta);//Like AllLightSequence but lights stay ON for 5.5 seconds (timer in function)  KEEP
  FillInOutFor(3, Green, 4); //fills in like FillIn, but reverses after in forward direction  Del(3) is interesting. rep(4)
  FillInOutRev(10,Red,4);//fills in like FillIn, but reverses after in opposite direction
  Fragments();// Uses arrays in non-convential manner to get a fragmented display, cycling all colors KEEP  exec 7 sec
  Nova(White, 4); // from center pentagon to outer tips of points.  timing changes? Pass color and # of repeats  KEEP
  NovaShimmer(Yellow);//Ok   Cool following Nova  may be too short.
  OrbitsCounterRot();//outer cycles CW in rgb, inner CCW in RGB maybe diff colors for pentagon. exec 10.6 sec
  OuterInner();//Basic but okay.  Outside lights in 6 colors (no Red), inside is Red, no center lights.  exec 10.9 sec 1 repeat only
  PentagonExpand(Cyan);// Pretty basic not sure how to use exec 0.5 sec (one loop). Single use, used by Nova
  PentagonCollapse(Red,3);// Only pentagon  exec 1.97 sec
  PentRotate2(20);// very good outer cycles 3 times, once for each RGB color, then inner does the same. at del(20) exec 4.0sec
  PointCollapse(20,4);//each point collapses from apex, 20-24, rgb colors all red, all green, then all blue,  4 reps exec 8.8sec
  DispVolts();
  PointExpand(10, Red, 4); // all points expand from base to apex in passed color  del(10) get simultaneous lighting, del(50)for rotation effec
  PointRotFill();//Takes 7 PointExpands at del 50 for rotation effect in 7 colors  KEEP  Exec 12.5 sec
  PointLineSweep(50,Magenta, 4);//Like pentagon, sweeps down 20, then 20+24, then 23+21  good at del(50), not good at (200) KEEP
  PointsCycleCW(10); //cycles all outside leds on points, in CW fashion, changing colors in process.  KEEP Pass delay value exec 7.8sec 
  Pulse(50,Red, 10);//like PentagonExpand(), but controllable timings.  Interesting running with PointLineSweep(Magenta) same timing
  Tips(500, 2); //lights just apex/corners of all segments.  Pass delay and reps. del(60)cool, del(2000) 
  FlashAll7Color(50);//Light each segment, CW rotation.  If followed by Rainbow, nice effect. runs for 3.56 seconds
  Rainbow(50, 8); //lights each complete segment one at a time, momentarily in diff color, pass delay and reps  Use with FlashAll7Color()
  RainbowAll(4000);// lights each segment in diff color and STAY ON for (x/1000)sec  Dimmer than normal. try to get all on bright
  RainbowCircArrays(50, 4);//pass Ontime value Like Rainbow() but circulates each of 7 colors CW. exec 12 sec for 4 repeats
  Resonance(2);// like OrbitsCounterRot, but same direction rotation.  cool  maybe diff colors for pentagons run with Orbits...cool
  Shimmer(6, 3);//all lights, all colors blinking very fast, like a "shiver". del(10) and 3 reps exec 3.2 sec
  Sparkle(4000); //For pentagons at 80-112, random flashing of all lights in 6 colors, pass time value KEEP cool! Don't put after Shimmer
  StaryNight(4000);//randomly lights any of 110 lights in different colors. exec 4 sec Very cool. KEEP
  Swarm();//pentagon only both all three rotate CW, good Like Orbits or Resonance???  exec 2.6 sec
  Tetris();//all colors, was PentagonCrawl() exec 10.1 sec
  FlashAll7Color(50);
  Resonance(2);//exec time 5.7 secs max power use very bright  Works well with FlashAll7Color() 
  Total(4);//every segment on sequentially, then blinks through all 7 colors Pass #of repeats
 
 // ExecTimer();//uncomment to run Timer.  Also top line.
     ////**** All functions: 255 secs or 4.25 minutes
  //AllTest(); //Test routine for all segments.   Could use as a display perhaps
  //SingTest(0);//Test routine for one segment, flash 3 colors.  Enter segment #0-6*
}
//****************************************************************************
//********************  Functions  *******************************************
//****************************************************************************
void AllEdges1()//formally PentagonEdges, now all segment edges. 4 light combos flicker in each segment, change color
{ Serial.println ("Function: AllEdges1"); 
  for (byte color = 0; color < 7; color++)
  { PORTB = rgb[color];
    for (byte i = 0; i < 5; i++)
    { for (segment = 0; segment < 7; segment++)//all segments including center
      { CS.portWrite(segment, SideArray[i]);//SideArray was only for Pentagon, but used here for effect with Points
        delay(20);
      }
    }
  }
}
//*************************************************************************
void AllOnSeq(byte color)// Lights each segment in CW sequence and stay on
{Serial.println ("Function: AllOnSeq"); 
  { PORTB = color; //cycle colors on pins 9,10,11
    for (byte segment = 0; segment < 7; segment++)
    { CS.portWrite (segment, allon);//No inversion ~
      delay (100);//light all in 1/10th second
    }
    delay (1000);  //keep lights on for 1 second
    for (byte segment = 0; segment < 7; segment++)
    { CS.portWrite (segment, off);//turn them all off
      delay (500);
    }
  }
}
//***************************************************************************
void AllLightSequence(int del)//light each light individually in diff colors, 0-111 sequentially
{ Serial.println ("Function: AllLightSequence"); 
  for (byte i = 3; i < 6; i++)//No inversion ~  Magenta, yellow and cyan
  { PORTB = rgb[i];
    for (int j = 0; j < 112; j++)
    { CS.digitalWrite(j, HIGH);
      delay(del);
      CS.digitalWrite(j, LOW);
      delay(del);
    }
  }
}
void BlackHole(byte color, byte rep)//Points and Pentagon shrink to center.  Idea: Nova after, all white
{ Serial.println ("Function: Blackhole"); 
  while (rep > 0) { //number of times to run this function before exit (rep)
    PORTB = color; //No inversion ~
    for (byte i = 0; i < 7; i++)
      for (byte segment = 0; segment < 5; segment++) //points only with this array
      { CS.portWrite(segment, PointsPairsArray[i]);
        delay(10);
        CS.portWrite(segment, off);
      }
    PORTB = Yellow;
    CS.portWrite(5, BlackHoleArray[0]);
    delay(15);
    CS.portWrite(5, off); //
    PORTB = White;
    CS.portWrite(6, BlackHoleArray[1]);
    delay(15);
    CS.portWrite(6, off); //

    CS.portWrite(6, BlackHoleArray[2]);
    delay(15);
    CS.portWrite(6, off);

    delay (250);
    rep --;
  } delay (1000); //delay here used for testing, may not need
}
//**************************************************************************
void BlinkSpinAccel()//points only, each lights up completely, CW rotation, accelerate, then change color
{ Serial.println ("Function: BlinkSpinAccel"); 
  for (byte color = 0; color < 4; color++)
  { PORTB = rgb[color];
    for (byte del = 3; del < 10; del++) //just do last 7 delay values in new DelayArrayFast array.

      for (byte segment = 0; segment < 5; segment++)
      { CS.portWrite(segment, allon);
        delay(DelayArrayFast[del]);
        CS.portWrite(segment, off);
        delay(DelayArrayFast[del]);
      }
  }
}
//****************************************************************
void Breathe2(byte segment, byte colpin, byte rep)
{ Serial.println ("Function: Breathe"); 
  //byte rep = 3;
  while (rep > 0)
  { CS.portWrite(segment, allon);
    for (int i = 180; i < 360; i++)//can start with i=0 or 180(faster)
    { //sin fades in and out ("breathe"), cos starts bright, fades out   See FADEOUT()
      analogWrite(colpin, (sin(i * 0.0174533) + 1) * 127);//Best one so far
      delay(10);//repetition rate. larger=slower
    }  delay (250);
    rep--;
    CS.portWrite(segment, off);
  }
}
//****************************************************************************
void FadeAll(byte color, byte rep)//fade all 110 lights.  Uses max power.  blinks on red or magenta fades to other.
{ Serial.println ("Function: FadeAll"); 
//rep = 4;
  while (rep > 0)
  {
    PORTB = color; //color here will MIXES with color in FadeInOut function.
    for (byte segment = 0; segment < 7; segment++)
    {
      CS.portWrite (segment, allon);
    }
    FadeOut(3, BluePin); //pass delay value "3" to FadeOut
    delay (1000);
    for (byte segment = 0; segment < 7; segment++)
    {
      CS.portWrite (segment, off);
    }
    rep--;
  }

  delay (1000);
}
//***************************************************************************
void FadeAllPoints(byte color, byte rep)//Fades all 5 points in and out in various colors then repeating.  Keep
{Serial.println ("Function: FadeAllPoints"); 
  while (rep > 0)
  { for (byte color = 0; color < 4; color++)//with color changes, quite cool: colors change in middle of fade.  Keep
    { PORTB = rgb[color];
      for (byte segment = 0; segment < 5; segment++)
      {
        CS.portWrite(segment, allon);

      }
      FadeInOut(2, RedPin); //whatever color chosen above mixes with this color, passed to the FadeInOut() function

      //CS.portWrite(segment, off);//if commented out, each segment lights and stays on
    } rep--;
  }
}

//***************************************************************************
void FadePentagons(byte color, byte rep)//color from Loop call mixes with color given below for FadeInOut
{ Serial.println ("Function: FadePentagons"); 
  while (rep > 0)
  { PORTB = color;//fade not as obvious if colors are mixed
    { CS.portWrite(5, allon);
      CS.portWrite(6, allon);//turn on all pentagon (5&6)
      FadeInOut(3, RedPin);
    }
    rep--;
    CS.portWrite(5, off);
    CS.portWrite(6, off);
  }
}
//***************************************************************************
void FillIn(int del, byte color)//light each light 0-111 sequentially, passed color, then all stay on if commented out (see below)
{ Serial.println ("Function: Fillin"); 
  int runtime = 5500; //on for 5.5 seconds
    long unsigned startTime = millis();
    //interval = runtime; // 10 seconds
    while (millis() - startTime < runtime)
    { PORTB = color;
      for (int j = 0; j < 112; j++)
      { CS.digitalWrite(j, HIGH);//No inversion ~
        delay(del);
        //CS.digitalWrite(j, LOW); //comment out to have lights STAY ON
      }
    }
  
}
//***************************************************************************
void FillInOutFor(int del, byte color, byte rep)//shut off each light 0-111 sequentially AFTER lighting them in same forward direction
{ Serial.println ("Function: FillInOutFor"); 
while (rep > 0)
  { PORTB = color;
    for (int j = 0; j < 112; j++)
    { CS.digitalWrite(j, HIGH);
      delay(del);
    }
    for (int j = 0; j < 112; j++)
      //for (int j = 112; j > 0 ; j--)
    { CS.digitalWrite(j, LOW);
      delay(del);
    }
    rep--;
  }
}
//***************************************************************************
void FillInOutRev(int del, byte color, byte rep)//shut off each light 0-111 sequentially AFTER lighting them in REVERSE direction (backs out)
{ Serial.println ("Function: FillInOutRev"); 
while (rep > 0)
  { PORTB = color;
    for (int j = 0; j < 112; j++)
    { CS.digitalWrite(j, HIGH);
      delay(del);
    }
    for (int j = 112; j > 0 ; j--)
    { CS.digitalWrite(j, LOW);
      delay(del);
    }
    rep--;
  }
}

//*************************************************************************
void FlashAll7Color(int del)//Flash each segment in 7 different colors. CW rotation
{Serial.println ("Function: FlashAll7Color"); 
for (byte color = 0; color < 7; color++)
  { PORTB = rgb[color];
    for (byte segment = 0; segment < 5; segment++)
    { CS.portWrite (segment, 0xFFFF);
      delay (del);
      CS.portWrite (segment, 0x0000);
      delay (del);
    }
  }
}
//****************************************************************************
void Fragments()//formally PentagonEdges, now all segment edges  Use for testing trying to change color before segment
{ Serial.println ("Function: Fragments"); 
  for (byte color = 0; color < 7; color++)//fragmented display, hence name.  cool.
  { PORTB = rgb[color];
    for (byte i = 0; i < 7; i++)
    { for (segment = 0; segment < 7; segment++)
      { CS.portWrite(segment, InsideArray[i]);//SideArray was only for Pentagon, but used here for effect with Points
        delay(20);//SweepArray gives random-like patterns in full brightness, Inside array similar
      }
    }
  }for (segment = 0; segment < 7; segment++)
  {CS.portWrite(segment, off);}
}

//*************************************************************************
void Nova(byte color, byte rep)//Calls Pentagon Expand function then expands points Pass rep=number of repeats
{ Serial.println ("Function: Nova"); 
  while (rep > 0) { //number of times to run this function before exit (rep
    { PentagonExpand(color);//color defined in calling function
      PORTB = color;
      for (byte i = 7; i > 0; i--)
        for (byte segment = 0; segment < 5; segment++) //points only with this array
        { CS.portWrite(segment, PointsPairsArray[i]);
          delay(10);
          CS.portWrite(segment, off);
        }
    }
    rep--;
  }}
//*************************************************************************
void NovaShimmer(byte color)//Calls Pentagon Expand function then expands points
{ Serial.println ("Function: NovaShimmer"); 
  PentagonExpand(color);//like Nova but points "shimmmer" out instead of sequential shutoff
  PORTB = color;//only difference is ~ to invert PointsPairsArray values
  for (byte i = 7; i > 0; i--)
    for (byte segment = 0; segment < 5; segment++) //points only with this array

    { CS.portWrite(segment, ~PointsPairsArray[i]);
      delay(10);
      CS.portWrite(segment, off);
    }
}
//***********************************************************************************
void OrbitsCounterRot() //Outer and inner pentagon sequence 0-14 outer CW, inner CCW.
{ Serial.println ("Function: OrbitsCounterRot"); 
  for (byte j = 0; j < 7; j++)
  { PORTB = rgb[j];
    for (byte i = 0; i < 15; i++)
    {
      CS.portWrite(5, ForSeqArray[i]);//No inversion ~
      CS.portWrite(6, RevSeqArray[i]);
      delay(100);
    }
  }
}
//*************************************************************************
void OuterInner()//outer inner pentagon only
{ Serial.println ("Function: OuterInner"); 
  rep=1;
while (rep>0)
  {for (byte color = 1; color < 7; color++) //no Red
  { PORTB = rgb[color];
    for (byte i = 0; i < 7; i++) //Lights border of outer pentagon cycles through colors RGB
    { CS.portWrite(5, OutsideArray[i]);
      delay(100);
      CS.portWrite(5, off);
    }
    for (byte j = 0; j < 5; j++) //Lights border of middle pentagon. took out center pentagon (first 5 elements)
    { PORTB = Red; //Outside is 6 colors, inside is Red
      CS.portWrite(6, InsideArray[j]);
      delay(100);
      CS.portWrite(6, off);
    }
  }
  rep--;
}
}
//**************************************************************************
void PentagonCollapse(byte color, byte rep)
{Serial.println ("Function: PentagonCollapse"); 
  while (rep>0) 
{PORTB = color;
  CS.portWrite(5, BlackHoleArray[0]);
  delay(150);
  CS.portWrite(5, off); //

  CS.portWrite(6, BlackHoleArray[1]);
  delay(150);
  CS.portWrite(6, off); //

  CS.portWrite(6, BlackHoleArray[2]);
  delay(150);
  CS.portWrite(6, off);

  CS.digitalWrite(111, HIGH);
  delay (500);
  CS.digitalWrite(111, LOW);

  delay (1000);
  rep--;
}}
//*****************************************************************************
void PentagonExpand(byte color)//Opposite of PentagonCollapse.  Inner to outer.  KEEP, used by others
{  Serial.println ("Function: PentagonExpand"); 
int del = 150; //don't change..used by others, eg Nova()
  { PORTB = color;
    CS.portWrite(6, BlackHoleArray[2]); //Middle and center pentagon
    delay(del);
    CS.portWrite(6, off); //

    CS.portWrite(6, BlackHoleArray[1]);
    delay(del);
    CS.portWrite(6, off); //

    CS.portWrite(5, BlackHoleArray[0]); //outside pentagon
    delay(del);
    CS.portWrite(5, off);
  } 
}

//*************************************************************************
void PentRotate1()//pentagon only, device 25 and 26  not impressive  Don't USE!
{ Serial.println ("Function: PentRotate1"); 
  //Outer CW rotation followed by Inner CW rotation  All (30) red , all green, then all blue
  //Outer circulates slower than inner by factor 16:1 (all inner, while outer increments 1)
  // for (byte i = 0; i < 3 ; i++) //use i=7 for 7 colors, i=3 for 3
  { PORTB = Blue; //cycle colors on pins 9,10,11
    for (int j = 80; j <= 95; j++) //device 25
    {
      i2cdev(j, del);
      for (int j = 96; j <= 111 ; j++) //Device 26
        i2cdev(j, del);
    }
  }
}
//************************************************************************
void PentRotate2(int del)//inner and outer pentagons
{ Serial.println ("Function: PentRotate2"); 
  //Outer cycles for all 3 colors, then inner cycles for all 3 colors
  for (byte i = 0; i < 3 ; i++)
  { PORTB = rgb[i];
    for (int j = 80; j <= 95; j++)//device 25
    {
      i2cdev(j, del);
    }
  }
  for (byte i = 0; i < 3 ; i++)
  { PORTB = rgb[i];
    for (int j = 96; j <= 111 ; j++) //Device 26
    {
      i2cdev(j, del);
    }
  }
}
//******************************************************************************
void PointCollapse(int del, byte rep)//Each point collapses from top light down to base, circulates CW Devices 20-24, R, G, then B, repeating
{ Serial.println ("Function: PointCollapse"); 
  while (rep > 0)
  { for (byte i = 0; i < 3; i++) //test only 3 on model, may be able to use 7 on full project
    { PORTB = rgb[i]; //cycle colors on pins 9,10,11
      for (byte segment = 0; segment < 5; segment++) //points only with this array
        for (byte i = 0; i < 7; i++)
        { CS.portWrite(segment, PointsPairsArray[i]);//inversion using ~ has interesting effect
          delay(del);
          CS.portWrite(segment, off);
        }
    }
    rep--;
  }
}
//*************************************************************************
void PointExpand(int del, byte color, byte rep)//Opposite of PointCollapse  Keep delay low to imply simultaneity  Looks like it goes berserk.
{ Serial.println ("Function: PointExpand"); 
  while (rep > 0)
  { PORTB = color;
    for (byte i = 7; i > 0; i--)//sometimes 0 (apex) doesn't light
      for (byte segment = 0; segment < 5; segment++) //Segment addresses 20-24

      { CS.portWrite(segment, PointsPairsArray[i]);//inversion ~ not work well
        delay(del);
        CS.portWrite(segment, off);
      }
    rep--;
  }
}
//**************************************************************************
void PointRotFill()//exec 12.47 sec
{ Serial.println ("Function: PointRotFill"); 
  PointExpand(50, Red, 1); // all points expand from base to apex in passed color  del(10) get simultaneous lighting
  PointExpand(50, Blue, 1);//only want to run each one time.  1=#repeats
  PointExpand(50, Green, 1);
  PointExpand(50, Yellow, 1);
  PointExpand(50, Cyan, 1); //at del(50), get a rotation effect
  PointExpand(50, Magenta, 1);
  PointExpand(50, White, 1);
}
//***************************************************************************
void PointLineSweep(int del, byte color, byte rep) // Like pentagon sweep, but star-wide, starting at apex of 20
{ Serial.println ("Function: PointLineSweep"); 
  while (rep > 0)
  { PORTB = color;
    for (byte ArrElem = 0; ArrElem < 7; ArrElem++)//fetch array element 7 elements
    { CS.portWrite(0, SweepArray20[ArrElem]);// Addr 20 is 0 to Centipede
      delay(del);
      CS.portWrite(0, off);
    }
    for (byte ArrElem = 0; ArrElem < 5; ArrElem++)//fetch array element 5 elements
    { CS.portWrite(1, SweepArray2122[ArrElem]);//
      CS.portWrite(4, SweepArray2324[ArrElem]);//
      delay(del);
      CS.portWrite(1, off);
      CS.portWrite(4, off);
    }
    for (byte ArrElem = 0; ArrElem < 5; ArrElem++)
    { CS.portWrite(2, SweepArray2122[ArrElem]);//
      CS.portWrite(3, SweepArray2324[ArrElem]);//
      delay(del);
      CS.portWrite(2, off);
      CS.portWrite(3, off);
    }
    rep--;
  }
}
//*************************************************************************
void PointsCycleCW(int del)//for all points
{ Serial.println ("Function: PointsCycleCW"); 
for (byte i = 6; i > 0; i--)//j has to be 6 or there is a blank spot
  { PORTB = rgb[i]; //cycle 7 colors on pins 9,10,11

    for (int j = 11; j >= 0; j--) //dev 20 //forward order makes lights rotate CW.  Note j>=0 and j--
    {
      i2cdev(j, del);
    }
    for (int j = 27; j >= 16; j--) //device 21
    {
      i2cdev(j, del);
    }
    for (int j = 43; j >= 32; j--) //device 22
    {
      i2cdev(j, del);
    }
    for (int j = 59; j >= 48; j--) //device 23
    {
      i2cdev(j, del);
    }
    for (int j = 75; j >= 64 ; j--) //Device 24
    {
      i2cdev(j, del);
    }
  }
}
//*****************************************************************************
void PointsCycleCCW()//done with more code as PointsRot2.  Keep this one for reference.
{ Serial.println ("Function: PointsCycleCCW"); 
for (byte i = 0; i < 7; i++)
  { PORTB = rgb[i]; //cycle 7 colors on pins 9,10,11

    for (int j = 64; j <= 75; j++) //device 24
    {
      CS.digitalWrite(j, HIGH);
      delay(10);
      CS.digitalWrite(j, LOW);
      delay(10);
    }
    for (int j = 48; j <= 59; j++) //device 23
    {
      CS.digitalWrite(j, HIGH);
      delay(10);
      CS.digitalWrite(j, LOW);
      delay(10);
    }
    for (int j = 32; j <= 43; j++) //device 22
    {
      CS.digitalWrite(j, HIGH);
      delay(10);
      CS.digitalWrite(j, LOW);
      delay(10);
    }
    for (int j = 16; j <= 27; j++) //device 21
    {
      CS.digitalWrite(j, HIGH);
      delay(10);
      CS.digitalWrite(j, LOW);
      delay(10);
    }
    for (int j = 0; j <= 11 ; j++) //Device 20
    {
      CS.digitalWrite(j, HIGH);
      delay(10);
      CS.digitalWrite(j, LOW);
      delay(10);
    }
  }
}
//*****************************************************************************
void Pulse(int del, byte color, byte rep)//Like Pentagon Expand but different timings
{ Serial.println ("Function: Pulse"); 
  while (rep > 0)
  { PORTB = color;
    CS.digitalWrite(111, HIGH);
    delay(del);

    CS.portWrite(6, BlackHoleArray[2]); //Middle and center pentagon
    delay(del);
    //CS.portWrite(6, off);

    CS.portWrite(6, BlackHoleArray[1]);
    delay(del);
    CS.portWrite(6, off); //if commented, middle pentagon stays on, if uncommented, smoother pentagon expansion

    CS.portWrite(5, BlackHoleArray[0]); //outside pentagon
    delay(del);
    CS.portWrite(5, off);
    rep--;
  }
}
//***************************************************************
void Rainbow(int del, byte rep)
{Serial.println ("Function: Rainbow"); 
  while (rep > 0)
  { for (byte color = 0, segment = 0; color < 7, segment < 7; color++, segment++) //must put both variables in same line or no work.
    { PORTB = rgb[color];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }
    rep--;
  }
  delay (500); //pause to let other functions catch up
}
//*************************************************************************
void RainbowAll(int runtime)//light each segment diff color, keep all colors ON.  Dim
{ Serial.println ("Function: RainbowAll"); 
unsigned long startMillis = millis();
  while (millis() - startMillis < runtime)
  { for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++) //must put both variables in same line or no work.
    { PORTB = rgb[i];
      CS.portWrite(segment, allon);
      //delay (100); no delay
      CS.portWrite(segment, off);
      //delay (100); no delay
    }
  } delay (500);
}
//*************************************************************************
void RainbowCircArrays(int del, int rep)
{ Serial.println ("Function: RainbowCircArrays"); 
  //uses 7 different arrays to circulate colors CW.  Like Rainbow() and FlashAll7Color
  //but changes start color (all rotate right) on each pass
  while (rep > 0)
  { for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++)//must put both variables in same line or no work.
    { PORTB = ColorArray1[i];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }
    for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++)
    { PORTB = ColorArray2[i];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }
    for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++)
    { PORTB = ColorArray3[i];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }

    for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++)
    { PORTB = ColorArray4[i];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }

    for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++)
    { PORTB = ColorArray5[i];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }

    for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++)
    { PORTB = ColorArray6[i];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }

    for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++)
    { PORTB = ColorArray7[i];
      CS.portWrite(segment, allon);
      delay (del);
      CS.portWrite(segment, off);
    }
    rep--;
    //delay(500;// continuous rotation, no hesitation if commented out.
  }
}
//*************************************************************************
void Resonance(byte rep)// pentagons only
{ Serial.println ("Function: Resonance");
  while (rep > 0) //exec 9 sec for rep=2
  { for (byte j = 0; j < 3; j++)
    { PORTB = rgb[j]; //cycle colors on pins 9,10,11
      for (byte i = 0; i < 15; i++)   //top-down
      {
        CS.portWrite(5, ForSeqArray[i]);
        CS.portWrite(6, ForSeqArray[i]);
        delay(100);
      }
    } rep--;
  }
}
//*************************************************************************
void Shimmer(int del, int rep)//Flash all lights, all colors fast recycle time so lights "shimmer".  Use sparingly: lots of power required
{ Serial.println ("Function: Shimmer");
  while (rep > 0)
  { for (byte i = 0; i < 7; i++)
    { PORTB = rgb[i]; //cycle colors on pins 9,10,11
      for (byte j = 0; j < 7; j++)
      { CS.portWrite (j, allon);
        delay (del);
        CS.portWrite (j, off);
        delay (del);
      }
    } rep --;
  }
}
//***************************************************************************
void Sparkle(int runtime)//For pentagons at 80-112, random flashing of all lights in 6 colors
{ Serial.println ("Function: Sparkle");
unsigned long startTime = millis();
  while (millis() - startTime <= runtime)
    for (byte i = 0; i < 7; i++) //
    { PORTB = rgb[i]; //cycle colors on pins 9,10,11
      int j = random(80, 112); //these are lights 1-32 on device 25 and 26
      // i2cdev(j,del);
      { CS.digitalWrite(j, HIGH);
        delay(1);//10 seems best delay
        CS.digitalWrite(j, LOW);
        delay(1);
      }
    }
}
//*************************************************************************
void StaryNight(int runtime)//randomly lights any of 110 lights in different colors.  Very cool
{ Serial.println ("Function: StaryNight");
unsigned long startTime = millis();
  while (millis() - startTime <= runtime)
    for (byte i = 0; i < 7; i++) //
    { PORTB = rgb[i]; //cycle colors on pins 9,10,11
      int j = random(112);
      // i2cdev(j,del);
      { CS.digitalWrite(j, HIGH);
        delay(10);//10 seems best delay
        CS.digitalWrite(j, LOW);
        delay(10);
      }
    }
}
//**************************************************************************
void Swarm()//for pentagons only (inner/outer)
{ Serial.println ("Function: Swarm");
  for (byte k = 0; k < 3; k++)
  { PORTB = rgb[k];
    for (byte i = 0; i < 15; i++)
    { CS.portWrite(5, ForSeqArray[i]);//inversion using ~ not work
      delay(15); //looks cool fast (10), and in colors like a swarm
      for (byte k = 1; k < 4; k++)
      { PORTB = rgb[k];
        CS.portWrite(6, ForSeqArray[i]);
        delay(15);
      }
    } CS.portWrite(5, off);//clear port expanders
    CS.portWrite(6, off);
  }
}
//*************************************************************************
void Tetris()//Crawl from top to bottom, both pentagons Can be modified for
//sweep up, left and right, in RGB
{ Serial.println ("Function: Tetris");
for (byte k = 0; k < 7; k++)//if k<8 get "black" which serves as a good pause
  { PORTB = rgb[k];//all colors
    for (byte i = 0; i < 15; i++)   //top-down
      //for (byte i=15; i > 0; i--)   //bottom-up, misses top
    {
      CS.portWrite(PortArray[i], SweepArray[i]);//Don't use ~
      delay(100); //looks better without center light 0xC800.  Using 0x4800 (no center)
    }
  } CS.portWrite(5, off);
  CS.portWrite(6, off);
}
//*************************************************************************
void Tips(int del, byte rep) //lights "tips" of points and corners of pentagon, blinking each in red-blue  cool
{ Serial.println ("Function: Tips");
  while (rep > 0)
  { for (byte i = 7; i > 1; i--)
    { PORTB = rgb[i]; //cycle 7 colors on pins 9,10,11
      for (byte j = 0; j <= 9 ; j++)

        CS.digitalWrite(TipsArray[j], HIGH);//delay (100) displays all colors, 10 is fast but cool.
      delay(del);
      // CS.digitalWrite(TipsArray[j], LOW);
      delay(del);
    }
    rep--;
  }
}
//***************************************************************
void Total(byte rep)//every segment on,then blinks ALL 112 lights through all 7 colors
{ Serial.println ("Function: Total");
byte count = rep;
  while (count > 0) {
    for (byte i = 0, segment = 0; i < 7, segment < 7; i++, segment++) //must put both variables in same line or no work.
    { PORTB = rgb[i];
      CS.portWrite(segment, allon);
      delay (200);
      CS.portWrite(5, off);
      CS.portWrite(6, off);
      //delay (500);
    }
    count --;
  }
}
//*************************** add functions above this line *******************
//*************************** End of Main Programming *************************

//*****************************************************************************
//***********************Test and Utility Functions (eg i2cdev)****************
//*****************************************************************************
void AllTest()//test all lights
{ for (byte i = 0; i < 7; i++)//all colors 0-7 red, green, blue, magenta, yellow, cyan, white
  { PORTB = rgb[i]; //cycle colors on pins 9,10,11
    for (byte segment = 0; segment < 7; segment++)//all segments, 0-6
    { CS.portWrite (segment, allon);
      delay (250);
      CS.portWrite (segment, off);
      delay (100);
    }
  }
}

//**************************************************************
void BreatheReal(byte color)//doesn't work, incomplete  more realistic when it works
{
  { breathe_time = millis();
    float val = (exp(sin(millis() / 2000.0 * PI)) - 0.36787944) * 108.0;
    analogWrite(color, val);
  }
}
//*******************************************************************
void DispVolts()//Volt measurement and display
{ lcd.setBacklight(255);
  int value5v = analogRead(A0); //measure 5 volt supply
  int value12v = analogRead(A1); //measure 12 volt supply
  float involt5 = (value5v * 5.0) / 1023; //internal 5 volt reference
  float involt12 = (value12v * 5.0) / 1023;
  lcd.setCursor (0, 3);
  lcd.print ("Volts= ");//5 volt measurement
  lcd.print (involt5);
  lcd.setCursor (13, 3); //12 volt measurement
  lcd.print (involt12 * 11.4);
}//fudge factor to get actual value +/- 0.2 volts  uses 100/10 volt divider
//**************************************************************
void ExecTimer()//use to measure execution time of function or entire sketch
{ unsigned long currentMillis = millis();

  unsigned long execTime = currentMillis - previousMillis;
  execTime = execTime / 1000;
  lcd.setCursor (0, 2);
  lcd.print ("Exec =");
  lcd.setCursor (7, 2);
  lcd.print (execTime);
  lcd.setCursor (12, 2);
  lcd.print (" sec");
  previousMillis=currentMillis;
  }
//*************************************************************************
void FadeInOut(int del, byte colpin) //uses cosine function and floats to get a smoother fade in and out.
{
  float inVal = 5.712;

  for (inVal = 5.712; inVal < 11.995; inVal = inVal + 0.006)//def 5.712 and cos gives smooth fade IN.
  { out = cos(inVal) * 127.5 + 127.5;//scales value of -1/1 to 0-255
    delay(del);
    analogWrite(colpin, out);
  }
}
//************************************************
void FadeOut(int del, byte colpin)// Instant on, then fade out
{
  static float inVal = 5.712;//default was 4.712 (where out=0)
  float out;
  for (inVal = 5.712; inVal < 10.995; inVal = inVal + 0.00628)// 5.712 for "in" removes blip at beginning. 6.283+4.712=10.995.
    //6.283 is 2*pi radians (360 deg). Decrease increment value to speed up fade (fewer samples).
    //0.01 fades in/out like a nuclear plant warning light, 0.001 very slow
  { out = sin(inVal) * 127.5 + 127.5;//scales value of -1/1 to 0-255
    delay(del);
    analogWrite(colpin, out);
  }
}
//****************************************************************
void FadeOutCos(byte segment, byte colpin)
{ CS.portWrite(segment, allon);
  for (int i = 0; i < 180; i++)//can do it with i=0 or 180(faster)
  { //cos instead of sin: starts bright, fades OUT  see BREATHE()
    analogWrite(colpin, (cos(i * 0.0174533) + 1) * 127);//Best one so far
    delay(12);//repetition rate. larger=slower
  }  delay (250);
}
//***************************************************************************
void i2cdev(int j, int del) //function to write to the port expander pin number j
{ CS.digitalWrite(j, HIGH);//pass individual led (one of 112) parameter j
  delay(del);
  CS.digitalWrite(j, LOW);
  delay(del);
}/*  device 20 (0) 0-15
     device 21 (1) 16-31
     device 22 (2) 32-47
     device 23 (3) 48-63
     device 24 (4) 64-79
     device 25 (5) 80-95
     device 26 (6) 96-111 */
//*************************************************************************
void SingTest(byte segment)//test one segment at a time
{ for (byte i = 0; i < 3; i++)
  { PORTB = rgb[i]; //cycle colors on pins 9,10,11
    { CS.portWrite (segment, allon);//enter segment address from call
      delay (500);
      CS.portWrite (segment, off);
      delay (500);
    }
  }
}
//***************************************************************
//****************Testing below this line************************
